package tools.descartes.dml.mm.applicationlevel.functions.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.commons.math3.distribution.EnumeratedIntegerDistribution;
import org.apache.commons.math3.distribution.EnumeratedRealDistribution;

import tools.descartes.dml.mm.applicationlevel.functions.BoolSampleList;
import tools.descartes.dml.mm.applicationlevel.functions.BooleanLiteral;
import tools.descartes.dml.mm.applicationlevel.functions.BoxedPDF;
import tools.descartes.dml.mm.applicationlevel.functions.ContinuousSample;
import tools.descartes.dml.mm.applicationlevel.functions.DoubleLiteral;
import tools.descartes.dml.mm.applicationlevel.functions.DoubleSampleList;
import tools.descartes.dml.mm.applicationlevel.functions.EnumSampleList;
import tools.descartes.dml.mm.applicationlevel.functions.ExponentialDistribution;
import tools.descartes.dml.mm.applicationlevel.functions.IntLiteral;
import tools.descartes.dml.mm.applicationlevel.functions.IntSampleList;
import tools.descartes.dml.mm.applicationlevel.functions.NormalDistribution;
import tools.descartes.dml.mm.applicationlevel.functions.ProbabilityFunction;
import tools.descartes.dml.mm.applicationlevel.functions.ProbabilityMassFunction;
import tools.descartes.dml.mm.applicationlevel.functions.ReplayFile;
import tools.descartes.dml.mm.applicationlevel.functions.SampleList;

/**
 * Provides helper functions for ProbabilityFunctions, e.g., to compute their
 * mean value.
 * 
 * @author Fabian Brosig
 */
public class RandomVariableHelper {
	/*
	 * I know it is ugly...but apparently there is no other solution that works
	 * (With QVTO blackboxes, the probfn parameter seems to be explicitly casted so
	 * that method overloading does not work as expected.)
	 */
	public static Double getNumericalMean(ProbabilityFunction probfn) {
		if (probfn instanceof IntLiteral)
			return _getNumericalMean((IntLiteral) probfn);
		if (probfn instanceof DoubleLiteral)
			return _getNumericalMean((DoubleLiteral) probfn);
		if (probfn instanceof BooleanLiteral)
			return _getNumericalMean((BooleanLiteral) probfn);
		if (probfn instanceof ExponentialDistribution)
			return _getNumericalMean((ExponentialDistribution) probfn);
		if (probfn instanceof NormalDistribution)
			return _getNumericalMean((NormalDistribution) probfn);
		if (probfn instanceof BoxedPDF)
			return _getNumericalMean((BoxedPDF) probfn);
		if (probfn instanceof ProbabilityMassFunction)
			return _getNumericalMean((ProbabilityMassFunction) probfn);
		if (probfn instanceof ReplayFile)
			return _getNumericalMean((ReplayFile) probfn);
		return _getNumericalMean(probfn);
	}

	public static Double getNumericalVariance(ProbabilityFunction probfn) {
		if (probfn instanceof IntLiteral)
			return _getNumericalVariance((IntLiteral) probfn);
		if (probfn instanceof DoubleLiteral)
			return _getNumericalVariance((DoubleLiteral) probfn);
		if (probfn instanceof BooleanLiteral)
			return _getNumericalVariance((BooleanLiteral) probfn);
		if (probfn instanceof ExponentialDistribution)
			return _getNumericalVariance((ExponentialDistribution) probfn);
		if (probfn instanceof NormalDistribution)
			return _getNumericalVariance((NormalDistribution) probfn);
		if (probfn instanceof BoxedPDF)
			return _getNumericalVariance((BoxedPDF) probfn);
		if (probfn instanceof ProbabilityMassFunction)
			return _getNumericalVariance((ProbabilityMassFunction) probfn);
		if (probfn instanceof ReplayFile)
			return _getNumericalVariance((ReplayFile) probfn);
		return _getNumericalVariance(probfn);
	}

	private static Double _getNumericalVariance(ReplayFile replay) {
		double variance = 0;
		try {
			Collection<Double> samples = collectSamples(replay);
			double mean = _getNumericalMean(samples);
			variance = samples.parallelStream().mapToDouble(sample -> Math.pow(sample, 2)).average().getAsDouble()
					- Math.pow(mean, 2);
		} catch (IOException e) {
			throw new IllegalArgumentException(e.getMessage());
		}
		return variance;
	}

	private static Double _getNumericalVariance(EnumSampleList enumSl) {
		throw new UnsupportedOperationException("Enums do not have a numerical variance.");
	}

	private static Double _getNumericalVariance(DoubleSampleList doubleSl) {
		double[] singletons = new double[doubleSl.getItems().size()];
		double[] probabilities = new double[doubleSl.getItems().size()];
		for (int i = 0; i < doubleSl.getItems().size(); i++) {
			singletons[i] = doubleSl.getItems().get(i).getValue().doubleValue();
			probabilities[i] = doubleSl.getItems().get(i).getProbability().doubleValue();
		}
		EnumeratedRealDistribution dist = new EnumeratedRealDistribution(singletons, probabilities);
		return dist.getNumericalVariance();
	}

	private static Double _getNumericalVariance(BoolSampleList boolSl) {
		throw new UnsupportedOperationException("Booleans do not have a numerical variance.");
	}

	private static Double _getNumericalVariance(IntSampleList intSl) {
		int[] singletons = new int[intSl.getItems().size()];
		double[] probabilities = new double[intSl.getItems().size()];
		for (int i = 0; i < intSl.getItems().size(); i++) {
			singletons[i] = intSl.getItems().get(i).getValue().intValue();
			probabilities[i] = intSl.getItems().get(i).getProbability().doubleValue();
		}
		EnumeratedIntegerDistribution dist = new EnumeratedIntegerDistribution(singletons, probabilities);
		return dist.getNumericalVariance();
	}

	private static Double _getNumericalVariance(SampleList sl) {
		if (sl instanceof IntSampleList)
			return _getNumericalVariance((IntSampleList) sl);
		else if (sl instanceof BoolSampleList)
			return _getNumericalVariance((BoolSampleList) sl);
		else if (sl instanceof DoubleSampleList)
			return _getNumericalVariance((DoubleSampleList) sl);
		else if (sl instanceof EnumSampleList)
			return _getNumericalVariance((EnumSampleList) sl);
		throw new IllegalStateException("Unknown type of SampleList: " + sl.getClass().getName());
	}

	private static Double _getNumericalVariance(ProbabilityMassFunction pmf) {
		return _getNumericalVariance(pmf.getSamples());
	}

	private static Double _getNumericalVariance(BoxedPDF bpdf) {
		throw new UnsupportedOperationException("BoxedPDF variance is unsupported!");
	}

	private static Double _getNumericalVariance(NormalDistribution nd) {
		return Math.pow(nd.getSigma().doubleValue(), 2);
	}

	private static Double _getNumericalVariance(ExponentialDistribution expd) {
		double rate = expd.getRate().doubleValue();
		if (rate > 0) {
			return 1 / Math.pow(rate, 2);
		}
		throw new IllegalArgumentException("ExponentialDistribution rate must be greater than 0.");
	}

	private static Double _getNumericalVariance(BooleanLiteral l) {
		throw new UnsupportedOperationException("Booleans do not have a numerical variance.");
	}

	private static Double _getNumericalVariance(DoubleLiteral l) {
		return 0.0;
	}

	private static Double _getNumericalVariance(IntLiteral l) {
		return 0.0;
	}

	private static Double _getNumericalVariance(ProbabilityFunction probfn) {
		throw new IllegalStateException("Unknown type of ProbabilityFunction: " + probfn.getClass().getName());
	}

	private static Double _getNumericalMean(ReplayFile replay) {
		double mean = 0;
		try {
			mean = _getNumericalMean(collectSamples(replay));
		} catch (IOException e) {
			throw new IllegalArgumentException(e.getMessage());
		}
		return mean;
	}

	private static Double _getNumericalMean(ProbabilityMassFunction pmf) {
		return _getNumericalMean(pmf.getSamples());
	}

	private static Double _getNumericalMean(IntSampleList intsl) {
		int[] singletons = new int[intsl.getItems().size()];
		double[] probabilities = new double[intsl.getItems().size()];
		for (int i = 0; i < intsl.getItems().size(); i++) {
			singletons[i] = intsl.getItems().get(i).getValue().intValue();
			probabilities[i] = intsl.getItems().get(i).getProbability().doubleValue();
		}
		EnumeratedIntegerDistribution dist = new EnumeratedIntegerDistribution(singletons, probabilities);
		return dist.getNumericalMean();
	}

	private static Double _getNumericalMean(DoubleSampleList doublesl) {
		double[] singletons = new double[doublesl.getItems().size()];
		double[] probabilities = new double[doublesl.getItems().size()];
		for (int i = 0; i < doublesl.getItems().size(); i++) {
			singletons[i] = doublesl.getItems().get(i).getValue().doubleValue();
			probabilities[i] = doublesl.getItems().get(i).getProbability().doubleValue();
		}
		EnumeratedRealDistribution dist = new EnumeratedRealDistribution(singletons, probabilities);
		return dist.getNumericalMean();
	}

	private static Double _getNumericalMean(SampleList sl) {
		if (sl instanceof IntSampleList)
			return _getNumericalMean((IntSampleList) sl);
		else if (sl instanceof BoolSampleList)
			return _getNumericalMean((BoolSampleList) sl);
		else if (sl instanceof DoubleSampleList)
			return _getNumericalMean((DoubleSampleList) sl);
		else if (sl instanceof EnumSampleList)
			return _getNumericalMean((EnumSampleList) sl);
		throw new IllegalStateException("Unknown type of SampleList: " + sl.getClass().getName());
	}

	private static Double _getNumericalMean(EnumSampleList enumsl) {
		throw new UnsupportedOperationException("Enums do not have a numerical mean.");
	}

	private static Double _getNumericalMean(BoolSampleList enumsl) {
		throw new UnsupportedOperationException("Booleans do not have a numerical mean.");
	}

	private static Double _getNumericalMean(BoxedPDF bpdf) {
		double mean = 0;
		double lastValue = 0;
		for (ContinuousSample cs : bpdf.getSample()) {
			mean += cs.getProbability().doubleValue() * (lastValue + (cs.getValue().doubleValue() - lastValue) / 2);
			lastValue = cs.getValue().doubleValue();
		}
		return mean;
	}

	private static Double _getNumericalMean(NormalDistribution nd) {
		return nd.getMu().doubleValue();
	}

	private static Double _getNumericalMean(ExponentialDistribution expd) {
		double rate = expd.getRate().doubleValue();
		if (rate > 0) {
			return 1 / rate;
		}
		throw new IllegalArgumentException("ExponentialDistribution rate must be greater than 0.");
	}

	private static Double _getNumericalMean(BooleanLiteral l) {
		throw new UnsupportedOperationException("Booleans do not have a numerical mean.");
	}

	private static Double _getNumericalMean(DoubleLiteral l) {
		return l.getValue().doubleValue();
	}

	private static Double _getNumericalMean(IntLiteral l) {
		return l.getValue().doubleValue();
	}

	private static Double _getNumericalMean(ProbabilityFunction probfn) {
		throw new IllegalStateException("Unknown type of ProbabilityFunction: " + probfn.getClass().getName());
	}

	private static Collection<Double> collectSamples(ReplayFile replay) throws IOException {
		List<Double> samples = new ArrayList<Double>();
		BufferedReader br = new BufferedReader(new FileReader(new File(replay.getFilepath()).getAbsolutePath()));
		while (br.ready()) {
			samples.add(Double.parseDouble(br.readLine()));
		}
		br.close();
		return samples;
	}

	private static Double _getNumericalMean(Collection<Double> samples) {
		return samples.parallelStream().mapToDouble(sample -> sample).average().getAsDouble();
	}
}
